/*
    WaveCtrlCmd.cc

    Will Eighteen, 2001: will@opaline.demon.co.uk
*/

#include "WaveCtrlCmd.h"
#include "DWaveCmd.h"
#include "AppWindow.h"
#include <Xm/RowColumn.h>
#include <Xm/ToggleB.h>
#include <Xm/PushB.h>
#include <Xm/Label.h>
#include "OpResources.h"
#include <MessageDialogManager.h>
#include "MixwaveUtils.h"


WaveCtrlCmd::WaveCtrlCmd ( const char *name, int active ) : NoUndoCmd ( name, active )
{
	strncpy ( _name, name, sizeof ( _name ) );
	_name [ sizeof ( _name ) - 1 ] = 0;
	_dWave = 0;
	mixwaveResources->waveCtrlRealised ( False );
}


WaveCtrlCmd::~WaveCtrlCmd()
{
	if ( _dWave )
		delete _dWave;
}


void WaveCtrlCmd::doCommand()
{
	if ( mixwaveResources->waveCtrlRealised() == True )
		return;

	if ( (*wavePak)[0].pWave->first() == NULL )
		return;

	mixwaveResources->waveCtrlRealised ( True );

	char title[256];
	unsigned int n;
	unsigned int maxNameLength = 0;
	
	_utils = new MixwaveUtils;
	_arrayItems = wavePak->Count();

	for ( n = 0; n < _arrayItems; n++ )
		if ( strlen ( (*wavePak)[n].pWave->name() ) > maxNameLength )
			maxNameLength = strlen ( (*wavePak)[n].pWave->name() );

	n = 0;
	maxNameLength++;
	
	if ( maxNameLength < 5 )
		maxNameLength = 5;	// accomodate 'quit' at a minimum
		
	_arrayItems++;
	
	char strings [ _arrayItems ] [ maxNameLength ];
	char waveName [ strlen ( mixwaveResources->currentWave() ) + 1 ];

	for ( n = 0; n < ( _arrayItems - 1 ); n++ )
		strcpy ( ( char * ) &strings [ n ], (*wavePak)[n].pWave->name() );

	strcpy ( strings [ n ], "Quit" );

	strncpy ( title, "Current wave: ", sizeof ( title ) );
	title [ sizeof ( title ) - 1 ] = 0;

	for ( n = 0; n < wavePak->Count(); n++ )
	{
		if ( strncmp ( (*wavePak)[n].pWave->name(), mixwaveResources->currentWave(),
				strlen ( mixwaveResources->currentWave() ) ) == 0 )
		{
			strcpy ( waveName, _utils->fileFromPath ( mixwaveResources->currentWave() ) );
			strncat ( title, waveName, sizeof ( title ) - strlen ( title ) - 1 );
			title [ sizeof ( title ) - 1 ] = 0;
		}
	}

	if ( strlen ( waveName ) == 0 )
		mixwaveResources->currentWave ( "" );	// current wave not on 'loaded' list

	_dWave = new DWaveCmd ( "WaveControl", TRUE, FALSE,
				&WaveCtrlCmd::dWaveSelectionCallback,
				( void * ) this, title,
				_arrayItems, maxNameLength, ( char * ) &strings );

	_dWave->execute();
}


void WaveCtrlCmd::dWaveSelectionCallback ( void *clientData, unsigned int selectedItemNum, char *label )
{
	WaveCtrlCmd *obj = ( WaveCtrlCmd * ) clientData;

	if ( ( strcmp ( label, "Quit" ) == 0 ) || ( selectedItemNum == 0 ) )
	{
		mixwaveResources->waveCtrlRealised ( False );
		delete obj->_utils;
		delete obj->_dWave;
	}
	else
	{
		int count, found;

		// scan base list for returned name pak index
		for ( count = 0; count < wavePak->Count(); count++ )
		{
			found = 0;

			if ( strcmp ( label, obj->_utils->fileFromPath ( (*wavePak)[count].pWave->name() ) ) == 0 )
			{
				found = 1;
				count = wavePak->Count();
			}
		}

		if ( !found )
		{
			if ( mixwaveResources->enableMessagePopups() == True )
			{
				char msg[BUFSIZE];
				snprintf ( msg, sizeof ( msg ), "Error: %s\nFile not loaded.", label );
				msg [ sizeof ( msg ) - 1 ] = 0;
				messageDialogManager->post ( msg );
			}

			label[0] = 0;
		}
		else
			obj->waveSelection ( label );
	}
}


void WaveCtrlCmd::waveSelection ( char *name )
{
	// we need the full name back and can't pass a possibly invalid
	// wave number (menu is static, wavePak contents may change.
	int count, found;
	
	for ( count = 0; count < wavePak->Count(); count++ )
	{
		found = 0;

		if ( strcmp ( name, _utils->fileFromPath ( (*wavePak)[count].pWave->name() ) ) == 0 )
		{
			mixwaveResources->currentWave ( (*wavePak)[count].pWave->name() );
			found = 1;
			count = wavePak->Count();
		}
	}
		
	if ( !found )
		mixwaveResources->currentWave ( "" );
}
