/*
	GraphIntCmd.cc - display integer data in graphical format

	Will Eighteen, 2000-2006: will@opaline.demon.co.uk
	
	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

	See COPYING for full details.
*/

#include <X11/cursorfont.h>
#include <Xm/Protocols.h>
#include "GraphIntCmd.h"
#include "Application.h"
#include <stdlib.h>
#include <string.h>
#include <Xm/Label.h>
#include <Xm/RowColumn.h>
#include <Xm/TextF.h>
#include <Xm/PushB.h>
#include <Xm/DrawingA.h>
#include <Xm/Frame.h>
#include <Xm/ToggleB.h>
#include <MessageDialogManager.h>
#include <Xm/ScrollBar.h>


GraphIntCmd::GraphIntCmd ( char *name, int active, GraphIntCallback callback, int numDigits,
				int *dataPtr, long dataOffset, long numDataPoints,
				long maxDataPoints, unsigned graphNum, void *clientData ) : NoUndoCmd ( name, active )
{
	_callback = callback;
	_numDigits = numDigits;
	_clientData = clientData;
	_data = dataPtr;
	_field1Value = _field2Value = _field3Value = 0;
	_field4Value = _field5Value = _field6Value = 0;
	_numData = numDataPoints;	// current no, data points in array
	_maxData = maxDataPoints;	// max data points externally available
	_graphNumber = graphNum;
	_startData = dataOffset;
	_windowWidth = 500;	// just a hint to the management
	_windowHeight = 300;
	_selecting = 0;
	_reload = False;
	_reloading = False;
	_refreshView = False;
	_drawPoints = False;
	_drawing = False;
	_minY = 0;	// need to check supplied bounds fit into NUMDIGITS
	_maxY = 0;
	_dataChanged = 0;
	_zoomStatus = False;
	_closeDown = False;
	_startSelectionLock = False;
	_endSelectionLock = False;
	_oldStart = dataOffset;
	_oldEnd = maxDataPoints - 1;
	_callbackType = 0;		// set by ok and cancel
	_centreZero = False;		// if True, data is signed, use 1/2 windowHeight	
	_lockScale = False;
	_interpolate = True;
	strncpy ( _applicationName, ( char * ) application->applicationClass(), sizeof ( _applicationName ) );
	_applicationName [ sizeof ( _applicationName ) - 1 ] = 0;
	strncat ( _applicationName, "Viewer", sizeof ( _applicationName ) - strlen ( _applicationName ) - 1 );
	// or the popup inherits the size of the top-level window
	strncpy ( _name, name, sizeof ( _name ) );
	_name [ sizeof ( _name ) - 1 ] = 0;
	_autoClose = True;	// activate 'ok' and 'cancel' vs. external 'close'
	_popup = 0;
}


void GraphIntCmd::doCommand()
{
	int n;
	Arg args[8];
	Atom WM_DELETE_WINDOW;

	_popup = XtCreatePopupShell ( _applicationName, transientShellWidgetClass,
						application->baseWidget(), NULL, 0 );

	_rc0 = XtVaCreateManagedWidget ( "RowColumn0", xmRowColumnWidgetClass, _popup,
						XmNorientation, XmHORIZONTAL,
						XmNnumColumns, 2, NULL );

	_frame = XtVaCreateManagedWidget ( "frame", xmFrameWidgetClass, _rc0, NULL );

	_rca = XtVaCreateManagedWidget ( "RowColumnA", xmRowColumnWidgetClass, _frame,
						XmNentryAlignment, XmALIGNMENT_CENTER,
						XmNorientation, XmVERTICAL,
	 					XmNnumColumns, 2, NULL );

	_rcb = XtVaCreateManagedWidget ( "RowColumnB", xmRowColumnWidgetClass, _rc0,
						XmNentryAlignment, XmALIGNMENT_CENTER,
						XmNorientation, XmVERTICAL,
	 					XmNnumColumns, 1, NULL );

	// use the supplied class instance name for the label widget text
	_label = XtCreateManagedWidget ( _name, xmLabelWidgetClass, _rca, NULL, 0 );

	_frame = XtVaCreateManagedWidget ( "frame", xmFrameWidgetClass, _rca, NULL );

	_canvas = XtVaCreateManagedWidget (
				( String ) "canvas", xmDrawingAreaWidgetClass, _frame,
				( String ) "exposeProc", ( XtArgVal ) exposeCallback,
				( String ) "exposeProcData", ( XtArgVal ) ( XtPointer ) this,
				( String ) "resizeProc", ( XtArgVal ) resizeCallback,
				( String ) "resizeProcData", ( XtArgVal ) ( XtPointer ) this,
				( String ) XtNheight, ( XtArgVal ) _windowWidth,
				( String ) XtNwidth, ( XtArgVal ) _windowHeight,
				NULL );


	acquireColour ( 3 );	// we _need_ colour here!
	n = 0;

	XtSetArg ( args[n], XmNorientation, XmHORIZONTAL ); n++;
	XtSetArg ( args[n], XmNminimum, 0 ); n++;
	XtSetArg ( args[n], XmNmaximum, _maxData +_numData - 1 ); n++;	// allow for sliderSize
	XtSetArg ( args[n], XmNsliderSize, _numData * ( ( _maxData + _numData ) / _maxData ) ); n++;
	XtSetArg ( args[n], XmNpageIncrement, _numData - 1 ); n++;
	XtSetArg ( args[n], XmNvalue, _startData ); n++;
	XtSetArg ( args[n], XmNtroughColor, _pixel[0] ); n++;

	if ( _numData == _maxData )
	{
		XtSetArg ( args[n], XmNshowArrows, False );
		n++;
	}
/* Eventually LessTif might support this; up to 0.92.26 it doesn't. */
/*
	if ( _startData == 0 )
		XtSetArg ( args[n], XmNshowArrows, XmMAX_SIDE ); n++;
	else	
		if ( _numData == _maxData )
			XtSetArg ( args[n], XmNshowArrows, XmMIN_SIDE ); n++;
		else
			if ( _startData > 0 )
				XtSetArg ( args[n], XmNshowArrows, XmEACH_SIDE ); n++;
			else
				XtSetArg ( args[n], XmNshowArrows, XmNONE ); n++;
*/
	/* widget only allows sb half the loaded file size or we don't get max
	value returned on drag*/
	_scrollBar = XtCreateManagedWidget ( "scrollBar", xmScrollBarWidgetClass, _rca, args, n );

	// general pointer motion callback
	XtAddEventHandler ( _canvas, PointerMotionMask, False,
				pointerMotionCallback, ( XtPointer ) this );

	// middle mouse button (or both) pressed motion callback (unused)
//	XtAddEventHandler ( _canvas, Button2MotionMask, False,
//				button2MotionCallback, ( XtPointer ) this );

	// right button pressed motion callback
	XtAddEventHandler ( _canvas, Button3MotionMask, False,
				button3MotionCallback, ( XtPointer ) this );

	XtAddEventHandler ( _canvas, ButtonReleaseMask, False,
				mouseButtonReleaseCallback, ( XtPointer ) this );

	XtAddEventHandler ( _canvas, ButtonPressMask, False,
				mouseButtonPressCallback, ( XtPointer ) this );

	XtAddEventHandler ( _canvas, ExposureMask, False,
				exposeCallback, ( XtPointer ) this );

	XtAddCallback ( _scrollBar, XmNvalueChangedCallback,
				&GraphIntCmd::scrolledValueChangedCallback, ( XtPointer ) this );

	XtAddCallback ( _scrollBar, XmNincrementCallback,
				&GraphIntCmd::scrollIncrementCallback, ( XtPointer ) this );

	XtAddCallback ( _scrollBar, XmNdecrementCallback,
				&GraphIntCmd::scrollDecrementCallback, ( XtPointer ) this );

	// now create another rowcol widget on the first
	/* first widget with range and selected range */
	_rc1 = XtVaCreateManagedWidget ( "RowColumn1", xmRowColumnWidgetClass,
					_rca,
					XmNorientation, XmHORIZONTAL,
					XmNpacking, XmPACK_COLUMN,
					XmNentryAlignment, XmALIGNMENT_CENTER,										
					XmNnumColumns, 2, NULL );

	n = 0;

	XtSetArg ( args[n], XmNcolumns, _numDigits ); n++;
	XtSetArg ( args[n], XmNeditable, TRUE ); n++;
	XtSetArg ( args[n], XmNcursorPositionVisible, TRUE ); n++;
	XtSetArg ( args[n], XmNmaxLength, _numDigits ); n++;

	_label1 = XtCreateManagedWidget ( "label1", xmLabelWidgetClass, _rc1, NULL, 0 );
	_label2 = XtCreateManagedWidget ( "label2", xmLabelWidgetClass, _rc1, NULL, 0 );
	_label3 = XtCreateManagedWidget ( "label3", xmLabelWidgetClass, _rc1, NULL, 0 );
	_label4 = XtCreateManagedWidget ( "label4", xmLabelWidgetClass, _rc1, NULL, 0 );
	_label5 = XtCreateManagedWidget ( "label5", xmLabelWidgetClass, _rc1, NULL, 0 );
	_label6 = XtCreateManagedWidget ( "label6", xmLabelWidgetClass, _rc1, NULL, 0 );

	_field1 = XtCreateManagedWidget ( "field1", xmTextFieldWidgetClass, _rc1, args, n );
	_field2 = XtCreateManagedWidget ( "field2", xmTextFieldWidgetClass, _rc1, args, n );
	_field3 = XtCreateManagedWidget ( "field3", xmTextFieldWidgetClass, _rc1, args, n );
	_field4 = XtCreateManagedWidget ( "field4", xmTextFieldWidgetClass, _rc1, args, n );
	_field5 = XtCreateManagedWidget ( "field5", xmTextFieldWidgetClass, _rc1, args, n );
	_field6 = XtCreateManagedWidget ( "field6", xmTextFieldWidgetClass, _rc1, args, n );

	// Add callbacks for data retrieval from parent
	XtAddCallback ( _field1, XmNactivateCallback, &GraphIntCmd::reloadCallback, ( XtPointer ) this );
	XtAddCallback ( _field2, XmNactivateCallback, &GraphIntCmd::reloadCallback, ( XtPointer ) this );
	XtAddCallback ( _field3, XmNactivateCallback, &GraphIntCmd::reloadCallback, ( XtPointer ) this );
	XtAddCallback ( _field4, XmNactivateCallback, &GraphIntCmd::reloadCallback, ( XtPointer ) this );
	XtAddCallback ( _field5, XmNactivateCallback, &GraphIntCmd::reloadCallback, ( XtPointer ) this );

	// trigger reload on datum change - provide continuous feedback
	XtAddCallback ( _field5, XmNvalueChangedCallback, &GraphIntCmd::reloadCallback, ( XtPointer ) this );
	XtAddCallback ( _field6, XmNvalueChangedCallback, &GraphIntCmd::reloadCallback, ( XtPointer ) this );

	// add callbacks for changed values
	XtAddCallback ( _field1, XmNvalueChangedCallback,
			&GraphIntCmd::field1ChangedCallback, ( XtPointer ) this );
	XtAddCallback ( _field2, XmNvalueChangedCallback,
			&GraphIntCmd::field2ChangedCallback, ( XtPointer ) this );
	XtAddCallback ( _field3, XmNvalueChangedCallback,
			&GraphIntCmd::field3ChangedCallback, ( XtPointer ) this );
	XtAddCallback ( _field4, XmNvalueChangedCallback,
			&GraphIntCmd::field4ChangedCallback, ( XtPointer ) this );
	XtAddCallback ( _field5, XmNvalueChangedCallback,
			&GraphIntCmd::field5ChangedCallback, ( XtPointer ) this );

	// add callbacks for data entry in text widgets
	XtAddCallback ( _field1, XmNactivateCallback,
			&GraphIntCmd::field1EntryCallback, ( XtPointer )  this );
	XtAddCallback ( _field2, XmNactivateCallback,
			&GraphIntCmd::field2EntryCallback, ( XtPointer )  this );
	XtAddCallback ( _field3, XmNactivateCallback,
			&GraphIntCmd::field3EntryCallback, ( XtPointer )  this );
	XtAddCallback ( _field4, XmNactivateCallback,
			&GraphIntCmd::field4EntryCallback, ( XtPointer )  this );
	XtAddCallback ( _field5, XmNactivateCallback,
			&GraphIntCmd::field5EntryCallback, ( XtPointer )  this );
	XtAddCallback ( _field6, XmNactivateCallback,
			&GraphIntCmd::field6EntryCallback, ( XtPointer )  this );

	if ( _autoClose == True )
	{
		_ok = XtVaCreateManagedWidget ( "label8", xmPushButtonWidgetClass, _rcb, NULL );
		XtAddCallback ( _ok, XmNactivateCallback, &GraphIntCmd::okCallback, ( XtPointer ) this );

		_cancel = XtCreateManagedWidget ( "label9", xmPushButtonWidgetClass, _rcb, NULL, 0 );
	    	XtAddCallback ( _cancel, XmNactivateCallback, &GraphIntCmd::cancelCallback, ( XtPointer ) this );
	}

	_drawType = XtCreateManagedWidget ( "Line", xmPushButtonWidgetClass, _rcb, NULL, 0 );
	XtAddCallback ( _drawType, XmNactivateCallback, &GraphIntCmd::drawTypeCallback, ( XtPointer ) this );

	_zoom = XtCreateManagedWidget ( "Zoom in", xmPushButtonWidgetClass, _rcb, NULL, 0 );
	XtAddCallback ( _zoom, XmNactivateCallback, &GraphIntCmd::zoomCallback, ( XtPointer ) this );

	_viewAll = XtCreateManagedWidget ( "label10", xmPushButtonWidgetClass, _rcb, NULL, 0 );
	XtAddCallback ( _viewAll, XmNactivateCallback, &GraphIntCmd::viewAllCallback, ( XtPointer ) this );

	_next = XtCreateManagedWidget ( "label11", xmPushButtonWidgetClass, _rcb, NULL, 0 );
	XtAddCallback ( _next, XmNactivateCallback, &GraphIntCmd::nextCallback, ( XtPointer ) this );

	_prev = XtCreateManagedWidget ( "label12", xmPushButtonWidgetClass, _rcb, NULL, 0 );
	XtAddCallback ( _prev, XmNactivateCallback, &GraphIntCmd::prevCallback, ( XtPointer ) this );

	_expand = XtCreateManagedWidget ( "label14", xmPushButtonWidgetClass, _rcb, NULL, 0 );
	XtAddCallback ( _expand, XmNactivateCallback, &GraphIntCmd::expandCallback, ( XtPointer ) this );

	_selRange = XtCreateManagedWidget ( "label13", xmPushButtonWidgetClass, _rcb, NULL, 0 );
	XtAddCallback ( _selRange, XmNactivateCallback, &GraphIntCmd::selRangeCallback, ( XtPointer ) this );

	_label7 = XtCreateManagedWidget ( "label7", xmLabelWidgetClass, _rcb, NULL, 0 );
	_field7 = XtCreateManagedWidget ( "field7", xmTextFieldWidgetClass, _rcb, args, n );

	XtAddCallback ( _field7, XmNvalueChangedCallback,
			&GraphIntCmd::field7ChangedCallback, (XtPointer ) this );

	XtAddCallback ( _field7, XmNactivateCallback,
			&GraphIntCmd::field7EntryCallback, ( XtPointer )  this );

	n = 0;

	XtSetArg ( args[n], XmNtoggleMode, XmTOGGLE_BOOLEAN ); n++;
 
	_startSelLock = XtCreateManagedWidget ( "toggle1", xmToggleButtonWidgetClass,
						_rcb, args, n );

	_endSelLock = XtCreateManagedWidget ( "toggle2", xmToggleButtonWidgetClass,
						_rcb, args, n );

	_scaleLock = XtCreateManagedWidget ( "toggle4", xmToggleButtonWidgetClass,
						_rcb, args, n );

	_interp = XtCreateManagedWidget ( "toggle3", xmToggleButtonWidgetClass,
						_rcb, args, n );

	XtVaSetValues ( _interp, XmNset, XmSET, NULL );	// switched on by default
											
	XtAddCallback ( _startSelLock, XmNvalueChangedCallback,
			&GraphIntCmd::startLockChangedCallback,( XtPointer ) this );

	XtAddCallback ( _endSelLock, XmNvalueChangedCallback,
			&GraphIntCmd::endLockChangedCallback,( XtPointer ) this );

	XtAddCallback ( _scaleLock, XmNvalueChangedCallback,
			&GraphIntCmd::scaleLockCallback,( XtPointer ) this );

	XtAddCallback ( _interp, XmNvalueChangedCallback,
			&GraphIntCmd::interpolateCallback,( XtPointer ) this );

	changeField1 ( _startData );
	changeField2 ( _startData + _numData - 1 );
// default y scale s/b half value range if signed, otherwise 1/range
	changeField7 ( 1 );	// default y scale

	// cater for WM 'X' pressed
	WM_DELETE_WINDOW = XmInternAtom ( XtDisplay ( _popup ), "WM_DELETE_WINDOW", False );
	XmAddWMProtocols ( _popup, &WM_DELETE_WINDOW, 1 );
	XmAddWMProtocolCallback ( _popup, WM_DELETE_WINDOW, clientMessageProc, ( XtPointer ) this );

	XtPopup ( _popup, ( XtGrabKind ) XtGrabNone );
	XtManageChild ( _popup );
}


void GraphIntCmd::clientMessageProc ( Widget, XtPointer clientData, XtPointer cbs )
{
	GraphIntCmd *obj = ( GraphIntCmd * ) clientData;

	if ( obj->_reloading == True )
		return;

	if ( obj->_closeDown == True )	// 'ok' already activated
		return;

	if ( obj->_autoClose == False )
		return;

	obj->clientMessage( ( XmAnyCallbackStruct * ) cbs );
}


void GraphIntCmd::clientMessage( XmAnyCallbackStruct *cbs )
{
	if ( cbs->reason != XmCR_CANCEL && cbs->reason != XmCR_OK )
	{
		_closeDown = True;
		cancel();
	}
}


void GraphIntCmd::setDataSource ( int *pData, long numSamples, long maxDataPoints )
{	// returned by calling code
	int sliderSize;
	int maximum;
	Boolean showArrows;

	_data = pData;
	_numData = numSamples;
	_maxData = maxDataPoints;

	if ( ( numSamples == 0 ) || ( maxDataPoints == 0 ) )	
		cancel();	// no data in file
	else
	{
		if ( _numData == _maxData )
		{
			showArrows = False;
			sliderSize = _maxData;
			maximum = _maxData;
		}
		else
		{
			showArrows = True;
			sliderSize = _numData * ( ( _maxData + _numData ) / _maxData );
			maximum = _maxData + _numData - 1;
		}

		setScrollBar ( maximum, sliderSize, _numData - 1, showArrows );
	}
}


void GraphIntCmd::setDatumValueBounds ( long minY, long maxY )
{	// check data fits into allowed space _numDigits
	char buf[2*_numDigits+1];
	
	snprintf ( buf, sizeof ( buf ), "%ld", minY );
	buf [ sizeof ( buf ) - 1 ] = 0;
	
	sscanf ( buf, "%ld", &_minY );

	snprintf ( buf, sizeof ( buf ), "%ld", maxY );
	buf [ sizeof ( buf ) - 1 ] = 0;

	sscanf ( buf, "%ld", &_maxY );
}


void GraphIntCmd::scrollIncrementCallback ( Widget, XtPointer clientData, XtPointer callData )
{
	GraphIntCmd *obj = ( GraphIntCmd * ) clientData;
	
	XmScrollBarCallbackStruct *cb = ( XmScrollBarCallbackStruct * ) callData;
	int num = cb->value;
	
	if ( num < obj->_field2Value )
	{
		XtVaSetValues ( obj->_scrollBar, XmNvalue, obj->_field2Value, NULL );
		num = obj->_field2Value;
	}
		
	obj->scrolledValueChanged ( num );
}


void GraphIntCmd::scrollDecrementCallback ( Widget, XtPointer clientData, XtPointer callData )
{
	GraphIntCmd *obj = ( GraphIntCmd * ) clientData;
	
	XmScrollBarCallbackStruct *cb = ( XmScrollBarCallbackStruct * ) callData;
	int num = cb->value;

	if ( num > obj->_field1Value )
	{
		XtVaSetValues ( obj->_scrollBar, XmNvalue, obj->_field1Value, NULL );
		num = obj->_field1Value;
	}

	obj->scrolledValueChanged ( num );
}


void GraphIntCmd::scrolledValueChangedCallback ( Widget, XtPointer clientData, XtPointer callData )
{
	GraphIntCmd *obj = ( GraphIntCmd * ) clientData;
	
	XmScrollBarCallbackStruct *cb = ( XmScrollBarCallbackStruct * ) callData;
	int num = cb->value;

	obj->scrolledValueChanged ( num );
}


void GraphIntCmd::scrolledValueChanged ( int value )
{
	if ( ( value >= _startData ) && ( value <= ( _startData + _numData - 1 ) ) )
		XtVaSetValues ( _scrollBar, XmNtroughColor, _pixel[0], NULL );

	if ( _field3Value != _field4Value )
		XtVaSetValues ( _scrollBar, XmNtroughColor, _pixel[2], NULL );	
	
	if ( value > ( _maxData - 1 ) )	// sb maximum o/s valid range...
		value = _maxData - 1;
	
	if ( value < _startData )
	{
		_field5Value = _startData;	// call back with legal value array[0];

		if ( value < 0 )
		{
			changeField1 ( 0 );
			changeField2 ( _numData - 1 );
		}
		else
		{
			changeField1 ( value );
			changeField2 ( value + _numData - 1 );
		}
	}
	else
	{
		if ( value > ( _startData + _numData - 1 ) )
		{
			_field5Value = _field2Value;	// array[_numData-1]

			if ( value > ( _maxData - 1 ) )
			{
				changeField2 ( _maxData - 1 );
				changeField1 ( _maxData - _numData - 1 );
			}
			else
			{	
				changeField2 ( value );
				changeField1 ( value - ( _numData - 1 ) );
			}
		}
	}

	_reload = True;
	reload();
	_reload = False;

	changeField5 ( _startData );	// set to datum value at new start, array[0]
	changeField6 ( _data[0] );
	refreshView();
}


void GraphIntCmd::startLockChangedCallback ( Widget, XtPointer clientData, XtPointer )
{
	GraphIntCmd *obj = ( GraphIntCmd * ) clientData;
	
	if ( obj->_startSelectionLock == True )
	{
		obj->_startSelectionLock = False;
		XtVaSetValues ( obj->_field3, XmNeditable, True, NULL );
	}
	else
	{
		obj->_startSelectionLock = True;
		XtVaSetValues ( obj->_field3, XmNeditable, False, NULL );
	}
}


void GraphIntCmd::endLockChangedCallback ( Widget, XtPointer clientData, XtPointer )
{
	GraphIntCmd *obj = ( GraphIntCmd * ) clientData;
	
	if ( obj->_endSelectionLock == True )
	{
		XtVaSetValues ( obj->_field4, XmNeditable, True, NULL );
		obj->_endSelectionLock = False;
	}
	else
	{
		obj->_endSelectionLock = True;
		XtVaSetValues ( obj->_field4, XmNeditable, False, NULL );
	}
}


void GraphIntCmd::scaleLockCallback ( Widget, XtPointer clientData, XtPointer )
{
	GraphIntCmd *obj = ( GraphIntCmd * ) clientData;
	
	if ( obj->_lockScale == True )
	{
		obj->_lockScale = False;
		XtVaSetValues ( obj->_field7, XmNeditable, True, NULL );
	}
	else
	{
		obj->_lockScale = True;
		XtVaSetValues ( obj->_field7, XmNeditable, False, NULL );
	}
}


void GraphIntCmd::interpolateCallback ( Widget, XtPointer clientData, XtPointer )
{
	GraphIntCmd *obj = ( GraphIntCmd * ) clientData;
	
	if ( obj->_interpolate == True )
		obj->_interpolate = False;
	else
		obj->_interpolate = True;
}


void GraphIntCmd::reloadCallback ( Widget, XtPointer clientData, XtPointer )
{
	GraphIntCmd *obj = ( GraphIntCmd * ) clientData;

	if ( obj->_drawing == False )	// doesn't prevent user field change... fatal
	{
		obj->_reloading = True;
		obj->reload();
	}
}


void GraphIntCmd::reload()
{
	if ( _callback )
	{
		if ( _reload == True )
			_reloading = True;

		_callback ( this, _clientData, _field1Value, _field2Value, _field3Value,
					_field4Value, _field5Value, _field6Value, _reload, _graphNumber,
					_callbackType );

		_dataChanged = 0;	// change must be picked up by callback implementation

		if ( _reload == True )
		{
			long min;

			min = _field1Value;

			if ( _field2Value < min )
				min = _field2Value;
		
			if ( _field3Value != _field4Value )
			{
				if ( _field3Value < min )
					min = _field3Value;
	
				if ( _field4Value < min )
					min = _field4Value;
			}

			_startData = min;	// numData is set externally by call to setDataSource()
		}

		_reloading = False;
	}
}


void GraphIntCmd::okCallback ( Widget, XtPointer clientData, XtPointer )
{
	GraphIntCmd *obj = ( GraphIntCmd * ) clientData;

	obj->ok();
}


Boolean GraphIntCmd::ok ()
{
	if ( _reloading == True )
		return ( False );
		
	if ( _closeDown == True )
		return ( False );

	_closeDown = True;
	_reload = False;	// shouldn't need setting here, but avoids oversight
	_callbackType = 1;	// really done, not just data report back
	reload();		// need the callback to pass data on closedown
	releaseColour ( 3 );
	XtDestroyWidget ( _popup );
	return ( True );
}


void GraphIntCmd::cancelCallback ( Widget, XtPointer clientData, XtPointer )
{
	GraphIntCmd *obj = ( GraphIntCmd * ) clientData;

	if ( obj->_reloading == True )
		return;

	if ( obj->_closeDown == True )	// 'ok' already activated
		return;

	obj->_closeDown = True;
	obj->cancel();
}


void GraphIntCmd::cancel()
{
	_reload = False;
	_callbackType = 2;
	reload();
	releaseColour ( 3 );
	XtDestroyWidget ( _popup );
}


void GraphIntCmd::viewAllCallback ( Widget, XtPointer clientData, XtPointer )
{
	GraphIntCmd *obj = ( GraphIntCmd * ) clientData;

	if ( obj->_reloading == True )
		return;
	
	obj->viewAll();
}


void GraphIntCmd::viewAll()
{
	if ( ( _field1Value == 0 ) && ( _field2Value == ( _maxData - 1 ) ) )
	{
		char msg[32];
		strcpy ( msg, "Entire data already displayed." );
		messageDialogManager->post ( msg );
		return;
	}
		
	changeField1 ( 0 );
	changeField2 ( _maxData - 1 );
	
	if ( ( _field1Value < _startData ) || ( _field2Value > ( _numData -1 ) ) )
	{
		_reload = True;
		reload();
		_reload = False;
	}
		
	setZoomLabel ( True );
	refreshView();
}


void GraphIntCmd::drawTypeCallback ( Widget, XtPointer clientData, XtPointer )
{
	GraphIntCmd *obj = ( GraphIntCmd * ) clientData;

	if ( obj->_drawPoints == True )
	{
		obj->_drawPoints = False;
		XmString xmstr = XmStringCreateLocalized ( "Line" );
		XtVaSetValues ( obj->_drawType, XmNlabelString, xmstr,
		XmNrecomputeSize, false, NULL );
		XmStringFree ( xmstr );
	}
	else
	{
		obj->_drawPoints = True;
		XmString xmstr = XmStringCreateLocalized ( "Point" );
		XtVaSetValues ( obj->_drawType, XmNlabelString, xmstr,
		XmNrecomputeSize, false, NULL );
		XmStringFree ( xmstr );
	}

	obj->refreshView();
}


void GraphIntCmd::zoomCallback ( Widget, XtPointer clientData, XtPointer )
{
	GraphIntCmd *obj = ( GraphIntCmd * ) clientData;

	if ( ( obj->_reloading == True ) || ( obj->_closeDown == True ) )
		return;

	obj->zoom();
}


void GraphIntCmd::zoom ( void )
{	
	Boolean showArrows;
	
	if ( _zoomStatus == False )
	{	// zoom in
		if ( _field3Value == _field4Value )
		{
			char msg[30];
			strcpy ( msg, "No selection to zoom to." );
			messageDialogManager->post ( msg );
			_zoomStatus = True;	 // so it gets reset to what is now
		}
		else
		{	// if the selected range isn't in the current range, reload
			_oldStart = _field1Value;
			_oldEnd = _field2Value;

			if ( ( ( _field3Value < _field1Value ) || ( _field3Value > _field2Value ) )
				|| ( _field4Value < _field1Value ) || ( _field4Value > _field2Value ) )
			{
				_reload = True;
				reload();
				_reload = False;
			}

			if ( _numData == _maxData )
				showArrows = False;
			else
				showArrows = True;

			setScrollBar ( _maxData + _numData - 1,
							( _field4Value - _field3Value ) * ( ( _maxData + _numData ) / _maxData ),
							_field4Value - _field3Value, showArrows );
							
			XtVaSetValues ( _scrollBar, XmNtroughColor, _pixel[2], NULL );
		}
	}
	else
	{	// zoom out	
		if ( ( _oldStart >= _startData ) && ( _oldEnd <= ( _startData + _numData ) ) )
			_reload = False;
		else
			_reload = True;	// region requested outside current range

		changeField1 ( _oldStart );
		changeField2 ( _oldEnd );

		if ( _reload == True )
		{
			reload();
			_reload = False;
		}

		if ( _numData == _maxData )
			showArrows = False;
		else
			showArrows = True;	

		setScrollBar ( _maxData + _numData - 1,
						_numData * ( ( _maxData + _numData ) / _maxData ),
						_numData - 1, showArrows );
	}	

	setZoomLabel ( _zoomStatus );
	refreshView();
}


void GraphIntCmd::nextCallback ( Widget, XtPointer clientData, XtPointer )
{
	GraphIntCmd *obj = ( GraphIntCmd * ) clientData;

	if ( ( obj->_reloading == True ) || ( obj->_closeDown == True ) )
		return;

	obj->next();
}


void GraphIntCmd::next()
{
	long temp;

	if ( ( _field1Value == 0 ) && ( _field2Value == ( _maxData - 1 ) ) )
	{
		char msg[32];
		strcpy ( msg, "Entire data already displayed." );
		messageDialogManager->post ( msg );
		return;
	}

	if ( _field3Value == _field4Value )
		XtVaSetValues ( _scrollBar, XmNtroughColor, _pixel[0], NULL );

	temp = _field2Value - _field1Value;

	if ( (_field2Value + temp ) > ( _maxData - 1 ) )
		temp = _maxData - _field2Value - 1;

	changeField2 ( _field2Value + temp );
	changeField1 ( _field1Value + temp );

	if ( ( _field2Value >= ( _numData - 1 ) ) && ( _field2Value <= ( _maxData - 1 ) ) )
	{
		_reload = True;
		reload();
		_reload = False;
	}

	if ( _field5Value < _field1Value )
	{
		changeField5 ( _field1Value );
		changeField6 ( _field1Value - _startData );
	}
		
		
	if ( _field5Value > _field2Value )
	{
		changeField5 ( _field2Value );
		changeField6 ( _field2Value - _startData );
	}

	XtVaSetValues ( _scrollBar, XmNvalue, _field2Value, NULL );
	refreshView();
}


void GraphIntCmd::prevCallback ( Widget, XtPointer clientData, XtPointer )
{
	GraphIntCmd *obj = ( GraphIntCmd * ) clientData;

	if ( ( obj->_reloading == True ) || ( obj->_closeDown == True ) )
		return;

	obj->prev();
}


void GraphIntCmd::prev()
{
	long temp;
	
	if ( ( _field1Value == 0 ) && ( _field2Value == ( _maxData - 1 ) ) )
	{
		char msg[32];
		strcpy ( msg, "Entire data already displayed." );
		messageDialogManager->post ( msg );
		return;
	}

	if ( _field3Value == _field4Value )
		XtVaSetValues ( _scrollBar, XmNtroughColor, _pixel[0], NULL );

	temp = _field2Value - _field1Value;

	if ( ( _field1Value - temp ) < 0 )
		temp = _field1Value;
	else
		temp = _field2Value - _field1Value;

	changeField2 ( _field2Value - temp );
	changeField1 ( _field1Value - temp );

	if ( _field1Value < _startData )			
	{
		_reload = True;
		reload();
		_reload = False;
	}

	if ( _field5Value < _field1Value )
	{
		changeField5 ( _field1Value );
		changeField6 ( _field1Value - _startData );
	}
		
	if ( _field5Value > _field2Value )
	{
		changeField5 ( _field2Value );
		changeField6 ( _field2Value - _startData );
	}
	
	XtVaSetValues ( _scrollBar, XmNvalue, _field1Value, NULL );
	refreshView();
}


void GraphIntCmd::expandCallback ( Widget, XtPointer clientData, XtPointer )
{
	GraphIntCmd *obj = ( GraphIntCmd * ) clientData;

	obj->expand();
}


void GraphIntCmd::expand ( void )
{
	Boolean showArrows;
	
	if ( _field2Value != _field4Value )
		changeField2 ( _field4Value );
	
	if ( _field1Value < ( _field2Value - 4 ) )
		changeField1 ( _field1Value + ( ( _field2Value - _field1Value )>>1 ) );

	refreshView();

	if ( _numData == _maxData )
		showArrows = False;
	else
		showArrows = True;

	setScrollBar ( _maxData + _numData - 1, ( _field2Value - _field1Value ) * ( ( _maxData + _numData ) / _maxData ),
					( _field2Value -_field1Value ), showArrows );

	XtVaSetValues ( _scrollBar, XmNvalue, _field1Value, NULL );
}


void GraphIntCmd::selRangeCallback ( Widget, XtPointer clientData, XtPointer )
{
	GraphIntCmd* obj = ( GraphIntCmd * ) clientData;

	obj->selRange();
}


void GraphIntCmd::datumChangeOkCallback ( XtPointer clientData )
{
	GraphIntCmd* obj = ( GraphIntCmd * ) clientData;

	obj->datumChangeOk ( obj->_field5Value );
}


void GraphIntCmd::datumChangeCancelCallback ( XtPointer clientData )
{
	GraphIntCmd* obj = ( GraphIntCmd * ) clientData;

	// existing value is ok, change text field back to _field6Value
	obj->changeField6 ( obj->_field6Value );
}


void GraphIntCmd::field1ChangedCallback ( Widget, XtPointer clientData, XtPointer )
{
	GraphIntCmd *obj = ( GraphIntCmd * ) clientData;
	char *text;
		
	text = XmTextFieldGetString ( obj->_field1 );
			
	if ( strlen ( text ) > 0 )
      	obj->field1Changed ( text );
}


void GraphIntCmd::field2ChangedCallback ( Widget, XtPointer clientData, XtPointer )
{
	GraphIntCmd *obj = ( GraphIntCmd * ) clientData;
	char *text;
		
	text = XmTextFieldGetString ( obj->_field2 );
		
	if ( strlen ( text ) > 0 )
		obj->field2Changed ( text );
}


void GraphIntCmd::field3ChangedCallback ( Widget, XtPointer clientData, XtPointer )
{
    GraphIntCmd *obj = ( GraphIntCmd * ) clientData;
	char *text;
		
	text = XmTextFieldGetString ( obj->_field3 );
		
	if ( strlen ( text ) > 0 )
       	obj->field3Changed ( text );
}


void GraphIntCmd::field4ChangedCallback ( Widget, XtPointer clientData, XtPointer )
{
	GraphIntCmd *obj = ( GraphIntCmd * ) clientData;
	char *text;
		
	text = XmTextFieldGetString ( obj->_field4 );
      	
	if ( strlen ( text ) > 0 )
		obj->field4Changed ( text );
}


void GraphIntCmd::field5ChangedCallback ( Widget, XtPointer clientData, XtPointer )
{
	GraphIntCmd *obj = ( GraphIntCmd * ) clientData;
	char *text;
		
	text = XmTextFieldGetString ( obj->_field5 );
		
	if ( strlen ( text ) > 0 )
      	obj->field5Changed ( text );
}


void GraphIntCmd::field7ChangedCallback ( Widget, XtPointer clientData, XtPointer )
{
	GraphIntCmd *obj = ( GraphIntCmd * ) clientData;

	if ( obj->_lockScale == True )	// prevent text entry
		return;

	char *text;

	text = XmTextFieldGetString ( obj->_field7 );

	if ( strlen ( text ) > 0 )
		obj->field7Changed ( text );
}


void GraphIntCmd::field1EntryCallback ( Widget, XtPointer clientData, XtPointer )
{
	GraphIntCmd* obj = ( GraphIntCmd * ) clientData;
	char text[obj->_numDigits+1];

	snprintf ( text, sizeof ( text ), "%s", XmTextFieldGetString ( obj->_field1 ) );
	text [ sizeof ( text ) - 1 ] = 0;

	if ( strlen ( text ) == 0 )
		obj->changeField1 ( 0 );

	obj->field1Entry ( obj->_field1Value );
}


void GraphIntCmd::field2EntryCallback ( Widget, XtPointer clientData, XtPointer )
{
	GraphIntCmd* obj = ( GraphIntCmd * ) clientData;
	char text[obj->_numDigits+1];

	snprintf ( text, sizeof ( text ), "%s", XmTextFieldGetString ( obj->_field2 ) );
	text [ sizeof ( text ) - 1 ] = 0;

	if ( strlen ( text ) == 0 )
		obj->changeField2 ( 0 );

	obj->field2Entry ( obj->_field2Value );
}


void GraphIntCmd::field3EntryCallback ( Widget, XtPointer clientData, XtPointer )
{
	GraphIntCmd* obj = ( GraphIntCmd * ) clientData;
	char text[obj->_numDigits+1];

	snprintf ( text, sizeof ( text ), "%s", XmTextFieldGetString ( obj->_field3 ) );
	text [ sizeof ( text ) - 1 ] = 0;

	if ( strlen ( text ) == 0 )	// if not, the value already corresponds
		obj->changeField3 ( obj->_field1Value );	// start sel is set to start display

	obj->field3Entry ( obj->_field3Value );
}


void GraphIntCmd::field4EntryCallback ( Widget, XtPointer clientData, XtPointer )
{
	GraphIntCmd* obj = ( GraphIntCmd * ) clientData;
	char text[obj->_numDigits+1];

	snprintf ( text, sizeof ( text ), "%s", XmTextFieldGetString ( obj->_field4 ) );
	text [ sizeof ( text ) - 1 ] = 0;

	if ( strlen ( text ) == 0 )
		obj->changeField4 ( obj->_numData - 1 );	// set end sel to end display

	obj->field4Entry ( obj->_field4Value );
}


void GraphIntCmd::field5EntryCallback ( Widget, XtPointer clientData, XtPointer )
{
	GraphIntCmd* obj = ( GraphIntCmd * ) clientData;
	char text[obj->_numDigits+1];

	snprintf ( text, sizeof ( text ), "%s", XmTextFieldGetString ( obj->_field5 ) );
	text [ sizeof ( text ) - 1 ] = 0;

	if ( strlen ( text ) == 0 )
	{
		obj->changeField5 ( obj->_field1Value );
		return;	// we didn't specify a datum number
	}

	obj->field5Entry ( obj->_field5Value );
}


void GraphIntCmd::field6EntryCallback ( Widget, XtPointer clientData, XtPointer )
{
	GraphIntCmd* obj = ( GraphIntCmd * ) clientData;
	char msg[256];
	char title[128];
	char numText[obj->_numDigits+1];

	// this changes actual data. post dialog with callback to change
	snprintf ( numText, sizeof ( numText ), "%s", XmTextFieldGetString ( obj->_field6 ) );
	numText [ sizeof ( numText ) - 1 ] = 0;
	strncpy ( title, obj->_name, sizeof ( title ) );
	title [ sizeof ( title ) - 1 ] = 0;
	snprintf ( msg, sizeof ( msg ), "%s\n\nReally alter datum value from %ld to %s ?", title, obj->_field6Value, numText );
	msg [ sizeof ( msg ) - 1 ] = 0;

	messageDialogManager->post ( msg, ( XtPointer ) clientData,
					&GraphIntCmd::datumChangeOkCallback,
					&GraphIntCmd::datumChangeCancelCallback );
}


void GraphIntCmd::field7EntryCallback ( Widget, XtPointer clientData, XtPointer )
{
	GraphIntCmd* obj = ( GraphIntCmd * ) clientData;
	char text[obj->_numDigits+1];

	snprintf ( text, sizeof ( text ), "%s", XmTextFieldGetString ( obj->_field7 ) );
	text [ sizeof ( text ) - 1 ] = 0;
	
	if ( strlen ( text ) == 0 )
		obj->changeField7 ( 1 );	// set y scale to unity

	obj->field7Entry ( obj->_field7Value );
}


void GraphIntCmd::field1Entry ( long value )
{
	char msg[30];
	int doReload = 0;

	if ( value > ( _numData - 1 ) )
	{
		if ( value > _maxData - 1 )
		{
			value = _maxData - 1;
			changeField1 (value );
		}
			
		doReload = 1;
	}
	
	if ( value > _field2Value )
	{
		changeField2 ( _field1Value );
		sprintf ( msg, "No data to display in range." );
		messageDialogManager->post ( msg );		
	}

	if ( value < _startData )
		doReload = 1;

	if ( doReload == 1 )
	{
		_reload = True;
		reload();
		_reload = False;
	}

	refreshView();
}


void GraphIntCmd::field2Entry ( long value )
{
	char msg[30];
	int doReload = 0;

	if ( value > ( _numData - 1 ) )
	{
		if ( value > _maxData - 1 )
		{
			value = _maxData - 1;
			changeField2 ( value );
		}
		
		doReload = 1;
	}
	
	if ( value < _field1Value )
	{
		changeField1 ( value );
		sprintf ( msg, "No data to display in range." );
		messageDialogManager->post ( msg );
	}

	if ( value < _startData )
		doReload = 1;

	if ( doReload == 1 )
	{
		_reload = True;
		reload();
		_reload = False;
	}

	refreshView();
}


void GraphIntCmd::field3Entry ( long value )
{
	char msg[34];
	int doReload = 0;

	if ( value > ( _numData - 1 ) )
	{
		if ( value > _maxData - 1 )
		{
			value = _maxData - 1;
			changeField3 ( value );
		}
		
		doReload = 1;
	}

	if ( value > _field4Value )
	{
		changeField4 ( value );
		sprintf ( msg, "No data to display in selection." );
		messageDialogManager->post ( msg );
	}

	if ( value < _startData )
		doReload = 1;
		
	if ( doReload == 1 )
	{
		_reload = True;
		reload();
		_reload = False;
	}

	setZoomLabel ( True );
	refreshView();
}


void GraphIntCmd::field4Entry ( long value )
{
	char msg[34];
	int doReload = 0;

	if ( value > ( _numData - 1 ) )
	{
		if ( value > _maxData - 1 )
		{
			value = _maxData - 1;
			changeField4 ( value );
		}
		
		doReload = 1;
	}

	if ( value < _field3Value )
	{
		changeField3 ( value );
		sprintf ( msg, "No data to display in selection." );
		messageDialogManager->post ( msg );
	}
		
	if ( value < _startData )
		doReload = 1;
		
	if ( doReload == 1 )
	{
		_reload = True;
		reload();
		_reload = False;
	}

	refreshView();
}


void GraphIntCmd::field5Entry ( long value )
{
	// reload already handled by field5Changed to provide continuous feedback
	// to the external callback
	changeField5 ( value );
	changeField6 ( _data[ value -_startData ] );
	refreshView();
}


void GraphIntCmd::field7Entry ( float value )
{
	// no check on mag limits
	changeField7 ( value );
	refreshView();	
}


void GraphIntCmd::field1Changed ( char *text )
{
	long value;
	
	if ( !checkTextEntry ( text ) )	// no error returned
	{
		sscanf ( text, "%ld", &value );

		if ( value > ( _maxData - 1 ) )
		{
			changeField1 ( _maxData - 1 );
			changeField2 ( _maxData - 1 );

			_reload = True;
			reload();
			_reload = False;
			refreshView();
		}
		else
			_field1Value = value;
	}
}


void GraphIntCmd::field2Changed ( char *text )
{	// 'e' is reserved for 'end file'
	long value;
	char c;

	c = text[0];

	if ( c == 'e' )
	{
		changeField2 ( ( long ) _maxData - 1 );
		
		if ( _field2Value > ( _numData - 1 ) )
		{
			_reload = True;
			reload();
			_reload = False;
		}
	}
	else
	{
		if ( !checkTextEntry ( text ) )
		{
			sscanf ( text, "%ld", &value );

			if ( value > ( _maxData - 1 ) )
			{
				changeField2 ( _maxData - 1 );
				_reload = True;
				reload();
				_reload = False;
				refreshView();
			}
			else
				_field2Value = value;
		}
	}
}


void GraphIntCmd::field3Changed ( char *text )
{
	long value;

	if ( !checkTextEntry ( text ) )
	{
		sscanf ( text, "%ld", &value );

		if ( value > ( _maxData - 1 ) )
		{

			changeField3 ( _maxData - 1 );
			changeField4 ( _maxData - 1 );
			_reload = True;
			reload();
			_reload = False;
			refreshView();
		}
		else
			_field3Value = value;
	}
}


void GraphIntCmd::field4Changed ( char *text )
{	// 'e' is reserved for 'end file'
	long value;
	char c;

	c = text[0];

	if ( c == 'e' )
	{
		changeField4 ( ( long ) _maxData - 1 );
		
		if ( _field4Value > ( _numData - 1 ) )
		{
			_reload = True;
			reload();
			_reload = False;
		}
	}
	else
	{
		if ( !checkTextEntry ( text ) )
		{
			sscanf ( text, "%ld", &value );

			if ( value > ( _maxData - 1 ) )
			{
				changeField4 ( _maxData - 1 );
				_reload = True;
				reload();
				_reload = False;
				refreshView();
			}
			else
				_field4Value = value;
		}
	}
}


void GraphIntCmd::field5Changed ( char *text )
{
	// change only allowed inside existing view range
	long value;
	
	if ( !checkTextEntry ( text ) )
	{
		sscanf ( text, "%ld", &value );
		_field5Value = value;
		
		if ( value < _field1Value )
			_field5Value = _field1Value;	// keep callback value within array bounds
		
		if ( value > _field2Value )
			_field5Value = _field2Value;
	}	
}


void GraphIntCmd::field7Changed ( char *text )
{
	char numText[_numDigits+1];

	snprintf ( numText, sizeof ( numText ), "%s", XmTextFieldGetString ( _field7 ) );
	numText [ sizeof ( numText ) - 1 ] = 0;
	sscanf ( numText, "%f", &_field7Value );
}


void GraphIntCmd::datumChangeOk ( long datumNumber )
{
	char numText[_numDigits+1];
	long value;

	snprintf ( numText, sizeof ( numText ), "%s", XmTextFieldGetString ( _field6 ) );
	numText [ sizeof ( numText ) - 1 ] = 0;
	sscanf ( numText, "%ld", &value );

	// check datum value is within allowed limits
	if ( ( value >= _minY ) && ( value <= _maxY ) && ( _minY != _maxY ) )
	{
		_data[datumNumber-_startData] = value;
		changeField6 ( value );
		_dataChanged = 1;	// set for one change per loaded block; client should check before reloading data
	}
	else
	{
		changeField6 ( _field6Value );
		_dataChanged = 0;
	}

	refreshView();
}


int GraphIntCmd::checkTextEntry ( char *text )
{	// reject anything non-numeric in data fields
	int error = 0;
	int length = strlen ( text );
	char c = 0;

	for ( int count = 0; count < length; count++ )
	{
		c = text [ count ];

		if ( ( c < '0' ) || ( c > '9' ) )
		{
			char msg[26];
			sprintf ( msg, "Error: '%c' non-numeric.", c );
			messageDialogManager->post( msg );
			error = 1;
		}
	}

	return error;
}

/* changeField: methods to set text box string from numeric value */
void GraphIntCmd::changeField1 ( long value )
{
	char text[_numDigits+1];
	
	_field1Value = value;
	snprintf ( text, sizeof ( text ), "%ld", value );
	text [ sizeof ( text ) - 1 ] = 0;

	XmTextFieldSetString ( _field1, text );
}


void GraphIntCmd::changeField2 ( long value )
{
	char text[_numDigits+1];
	
	_field2Value = value;
	snprintf ( text, sizeof ( text ), "%ld", value );
	text [ sizeof ( text ) - 1 ] = 0;

	XmTextFieldSetString ( _field2, text );
}


void GraphIntCmd::changeField3 ( long value )
{
	if ( _startSelectionLock == True )
		return;

	char text[_numDigits+1];
	
	_field3Value = value;
	snprintf ( text, sizeof ( text ), "%ld", value );
	text [ sizeof ( text ) - 1 ] = 0;

	XmTextFieldSetString ( _field3, text );
}


void GraphIntCmd::changeField4 ( long value )
{
	if ( _endSelectionLock == True )
		return;

	char text[_numDigits+1];
	
	_field4Value = value;
	snprintf ( text, sizeof ( text ), "%ld", value );
	text [ sizeof ( text ) - 1 ] = 0;

	XmTextFieldSetString ( _field4, text );
}


void GraphIntCmd::changeField5 ( long value )
{
	char text[_numDigits+1];

	_field5Value = value;
	snprintf ( text, sizeof ( text ), "%ld", value );
	text [ sizeof ( text ) - 1 ] = 0;

	XmTextFieldSetString ( _field5, text );
}


void GraphIntCmd::changeField6 ( long value )
{
	char text[_numDigits+1];
	
	_field6Value = value;
	snprintf ( text, sizeof ( text ), "%ld", value );
	text [ sizeof ( text ) - 1 ] = 0;

	XmTextFieldSetString ( _field6, text );
}


void GraphIntCmd::changeField7 ( float value )
{
	char text[_numDigits+1];

	snprintf ( text, sizeof ( text ), "%f", value );
	text [ sizeof ( text ) - 1 ] = 0;
	_field7Value = value;	// internally stored to full accuracy

	XmTextFieldSetString ( _field7, text );
}


void GraphIntCmd::pointerMotionCallback ( Widget, XtPointer clientData, XEvent *event, Boolean * )
{
	GraphIntCmd *graphCmd = ( GraphIntCmd * ) clientData;

	if ( ( graphCmd->_reloading == True ) || ( graphCmd->_closeDown == True ) )
		return;
	
	int xpos, ypos;
	long offsetDatumNum;

	xpos = event->xmotion.x;
	ypos = event->xmotion.y;
	
	if ( ( graphCmd->_selecting ) && ( graphCmd->_lockScale == False ) )
		graphCmd->dragYScale ( ypos  );

	if ( ( graphCmd->_selecting ) && ( xpos < 0 ) )
	{
		offsetDatumNum = graphCmd->datumNum ( -xpos );

		// put start selection at new data start
		if ( ( graphCmd->_field1Value - offsetDatumNum ) < 0 )
		{
			graphCmd->changeField1 ( 0 );
			graphCmd->changeField3 ( 0 );
		}
		else
		{
			graphCmd->changeField1 ( graphCmd->_field1Value - offsetDatumNum );
			graphCmd->changeField3 ( graphCmd->_field1Value );
		}

		graphCmd->_reload = True;
		graphCmd->reload();
		graphCmd->_reload = False;
			
		// now the end is the same as the start so we don't draw anything
		// reset so it's at the end of the window
		// this is done in button1Motion, so set xCoord here
		graphCmd->_xCoord = graphCmd->_windowWidth - 1;
	}
	else
	{
		if ( ( graphCmd->_selecting ) && ( xpos > ( int ) graphCmd->_windowWidth ) )
		{
			if ( graphCmd->_numData <= graphCmd->_maxData )
			{
				graphCmd->_xCoord = xpos;
				offsetDatumNum = graphCmd->datumNum ( graphCmd->_xCoord );
				offsetDatumNum += graphCmd->_field1Value;

				if ( ( offsetDatumNum ) > ( graphCmd->_maxData - 1 ) )
					offsetDatumNum = graphCmd->_maxData - 1;

				graphCmd->changeField2 ( offsetDatumNum );

				graphCmd->_reload = True;
				graphCmd->reload();
				graphCmd->_reload = False;
				graphCmd->_xCoord = graphCmd->_windowWidth - 1;
			}
		}
		else	
		{	// drag within limits of loaded data
			graphCmd->_xCoord = event->xmotion.x;
    		graphCmd->_yCoord = event->xmotion.y;

			if ( graphCmd->_drawing == True )
			{
				if ( graphCmd->_xCoord < 0 )
					graphCmd->_xCoord = 0;
				else
					if ( graphCmd->_xCoord > ( ( int ) graphCmd->_windowWidth - 1 ) )
						graphCmd->_xCoord = graphCmd->_windowWidth - 1;
			}
		}
	}

	graphCmd->pointerMotion();
}


void GraphIntCmd::button3MotionCallback ( Widget, XtPointer clientData, XEvent *event, Boolean * )
{
	GraphIntCmd *graphCmd = ( GraphIntCmd * ) clientData;

	if ( ( graphCmd->_reloading == True ) || ( graphCmd->_closeDown == True ) )
		return;

	// right mouse button: draw new graph region
	graphCmd->button3Motion();
}


void GraphIntCmd::pointerMotion()
{
    if ( _selecting ) // redraw to dragged position whilst mouse button pressed
    {
		changeField4 ( datumNum ( _xCoord ) + _field1Value );

		if  ( ( _xCoord == 0 ) || ( _xCoord == ( ( int ) _windowWidth - 1 ) ) )
			refreshView();	// redisplay
		else
			redrawSelection ( _field3Value, _field4Value );
    }

	if ( _field5Value < _startData )
	{
		changeField5 ( _startData );
		changeField6 ( datumValue ( 0 ) );	// _startData = array[0]
	}
	else
	{
		if ( _field5Value > ( _startData + _numData - 1 ) )
		{
			changeField5 ( _startData + _numData - 1 );
			changeField6 ( datumValue ( _numData - 1 ) );
		}
		else
		{
			changeField5 ( datumNum ( _xCoord ) + _field1Value );
			changeField6 ( datumValue ( datumNum ( _xCoord ) + ( _field1Value - _startData ) ) );
		}

	}
	
	XtVaSetValues ( _scrollBar, XmNvalue, _field5Value, NULL );
}


void GraphIntCmd::button3Motion()
{	
	if ( _xCoord < _lastDrawStart )
		return;

	long endArrayRef = datumNum ( _xCoord ) + _field1Value;
	long count = datumNum ( _lastDrawStart ) + _field1Value;
	int yValue = _yCoord;
	int centreY;
	float datumValue;

	if ( _centreZero == True )
		centreY = _windowHeight / 2;
	else
		centreY = _windowHeight;	

	yValue = centreY - yValue;	// inverted

	if ( _centreZero == True )
		datumValue = ( ( float ) yValue / ( float ) centreY ) * ( float ) ( ( _maxY - _minY ) / 2 );
	else
		datumValue = ( ( float ) yValue / ( float ) centreY ) * ( float ) ( _maxY - _minY );

	datumValue /= _field7Value;

	if ( datumValue > _maxY )
		datumValue = _maxY;

	if ( datumValue < _minY )
		datumValue = _minY;

	while ( count <= endArrayRef )
	{
		_data [ count - _startData ] = ( int ) datumValue;
		count++;
	}

	changeField3 ( datumNum ( _lastDrawStart ) + _field1Value );
	changeField4 ( datumNum ( _xCoord ) + _field1Value );
	redrawSelection ( _field3Value, _field4Value );
	_lastDrawStart = _xCoord;
}


void GraphIntCmd::dragYScale ( int ypos )
{
	int margin = _windowHeight / 4;

	ypos -= _windowHeight / 2;
		
	if ( ypos > margin )
		changeField7 ( ( 1 / ( float ) ( ypos - margin ) ) );
	else
	{
		if ( ypos < -margin )
			changeField7 ( ( float ) ( -margin - ypos ) );
		else
			changeField7 ( 1 );
	}
	
	refreshView();
}


void GraphIntCmd::mouseButtonPressCallback ( Widget, XtPointer clientData, XEvent *event, Boolean *)
{
	GraphIntCmd *graphCmd = ( GraphIntCmd * ) clientData;

	if ( ( graphCmd->_reloading == True ) || ( graphCmd->_closeDown == True ) )
		return;

	switch ( event->xbutton.button )	// action on button<n> pressed
	{
		case 1:
				graphCmd->mouseButton1Press();
				break;
		case 3:
				graphCmd->_lastDrawStart = event->xmotion.x;
				graphCmd->mouseButton3Press();
				break;
		default:
				return;
	}		
}


void GraphIntCmd::mouseButton1Press()
{
	changeField3 ( datumNum ( _xCoord ) + _field1Value );
	changeField4 ( datumNum ( _xCoord ) + _field1Value );

	if ( _zoomStatus == True )	// already zoomed in
		refreshView();			// reset selection colour so it's not bg colour
	else
	{
		_refreshView = True;
		redrawSelection ( _field3Value, _field4Value );
	}
	
	_selecting = 1;
	_zoomStatus = False;	// haven't selected yet

	XmString xmstr = XmStringCreateLocalized ( "Zoom in" );
	XtVaSetValues ( _zoom, XmNlabelString, xmstr, XmNrecomputeSize, false, NULL );
	XmStringFree ( xmstr );

	XtVaSetValues ( _scrollBar, XmNtroughColor, _pixel[2], NULL );
}


void GraphIntCmd::mouseButton3Press()
{
	_drawing = True;
	changeField3 ( datumNum ( _xCoord ) + _field1Value );
	changeField4 ( datumNum ( _xCoord ) + _field1Value );
}


void GraphIntCmd::mouseButtonReleaseCallback ( Widget, XtPointer clientData, XEvent *event, Boolean *)
{
	GraphIntCmd *graphCmd = ( GraphIntCmd * ) clientData;

	if ( ( graphCmd->_reloading == True ) || ( graphCmd->_closeDown == True ) )
		return;

	graphCmd->mouseButtonRelease ( event->xbutton.button );
}


void GraphIntCmd::mouseButtonRelease ( int buttonReleased )
{
	switch ( buttonReleased )
	{
		case 1:
				_selecting = 0;    // turn off redrawSel in button1Motion
				break;
		case 3:
				_dataChanged = True;
				_drawing = False;
				break;
		default:
				break;
	}
	
	changeField4 ( datumNum ( _xCoord ) + _field1Value );
	checkSelection();
	_reload = False;
	reload();	// report back if selection made via mouse

	if ( _numData < _maxData )
		XtVaSetValues ( _scrollBar, XmNshowArrows, True, NULL );

	if ( _field3Value != _field4Value )
		XtVaSetValues ( _scrollBar, XmNtroughColor, _pixel[2], NULL );

	refreshView();	// redisplay selection if pointer dragged left before release
}


void GraphIntCmd::exposeCallback ( Widget, XtPointer clientData, XEvent *event, Boolean *)
{
	GraphIntCmd *graphCmd = ( GraphIntCmd * ) clientData;

	graphCmd->redraw();
}


void GraphIntCmd::checkSelection()
{
	if ( _field3Value > _field4Value )	// we dragged right-to-left, swap
	{	
		long temp = _field4Value;
		changeField4 ( _field3Value );
		changeField3 ( temp );
	}
}


long GraphIntCmd::datumNum ( long xPosition )
{	// rel datum num in array w/respect to start of window (always datum 0)
	long datumNum;
	float windowRatio;

	windowRatio = (float) xPosition / (float) (_windowWidth - 1);
	datumNum = (long ) ( windowRatio * ( _field2Value - _field1Value ) );

	return datumNum;
}


long GraphIntCmd::datumValue ( long datumNum )
{	// relative datum num in array, not abs position in data
	long value;

	value = _data [ datumNum ];
	return value;
}
	

void GraphIntCmd::resizeCallback ( Widget w, XtPointer clientData )
{
	GraphIntCmd *graphCmd = ( GraphIntCmd *) clientData;

	graphCmd->getGeometry();
	graphCmd->redraw();
}


void GraphIntCmd::getGeometry ( void )
{
    int scrapI;
    unsigned scrapU;
    Window rootReturn;

	XGetGeometry( XtDisplay ( _canvas ), XtWindow ( _canvas ),
					&rootReturn,
					&scrapI, &scrapI,
					&_windowWidth, &_windowHeight,
					&scrapU, &scrapU );
}


void GraphIntCmd::refreshView()
{
	_refreshView = True;
	redraw();
	_refreshView = False;
	redraw();
}


void GraphIntCmd::redraw ( void )
{
	if ( _numData == 0 )
		return; // no data to plot

  	if ( !XtIsRealized ( _canvas ) ) return;

  	GC _graphicsContext;
	long pX,pY;
	int  centreY, screenX;
	double scaleX,scaleY;
	int oldX, oldY;
		
	setBusyCursor();
	getGeometry();

  	_graphicsContext = XCreateGC ( XtDisplay ( _canvas ),
					DefaultRootWindow ( XtDisplay (_canvas ) ), 0, NULL ); 

	if ( _refreshView == TRUE )
	{
		XSetForeground ( XtDisplay ( _canvas ), _graphicsContext, _pixel[0] );
		XFillRectangle ( XtDisplay ( _canvas ), XtWindow ( _canvas ),
							_graphicsContext, 0,0,_windowWidth, _windowHeight);
		return;
	}
	else
	{
		XSetForeground ( XtDisplay ( _canvas ), _graphicsContext, _pixel[2] );
		scaleX = ( float ) ( _windowWidth - 1 ) / ( float ) ( _field2Value - _field1Value );

		if ( _centreZero == True )
		{
			centreY = _windowHeight / 2;
			scaleY = ( float )  ( _windowHeight ) / ( float ) ( _minY - _maxY );
		}
		else
		{
			centreY = _windowHeight;
			scaleY = - ( float ) _windowHeight / ( float ) ( _maxY - _minY );
		}

		scaleY *= _field7Value;	// user-defineable mag override

		// initialise first datum value	
		pY = ( long ) ( ( int ) ( centreY + ( _data[ _field1Value - _startData ] * scaleY ) ) );

		oldX = 0;
		oldY = pY;
		int increment = 1;	// default not interpolation

		if ( _interpolate == True )
		{
			increment = ( _field2Value - _field1Value ) / _windowWidth;
			
			if ( increment == 0 )
				increment = 1;
		}

		for ( pX = _field1Value; pX <= _field2Value; pX += increment )
		{
			screenX = ( int ) ( ( pX - _field1Value ) * scaleX );	// x position of screen
			pY = ( long ) ( ( int ) ( centreY + ( _data[ pX - _startData ] * scaleY ) ) );

			if ( _drawPoints == True )
			{
				oldX = screenX;
				oldY = pY;
			}
	
			XDrawLine( XtDisplay ( _canvas ), XtWindow ( _canvas ),
						_graphicsContext, oldX, oldY, screenX, pY );

			oldX = screenX;
			oldY = pY;
		}

		redrawSelection ( _field3Value, _field4Value );
	}

	setNormalCursor();
}


void GraphIntCmd::redrawSelection ( long startDrawDatumNum, long endDrawDatumNum )
{
	// 'data per screen x unit' is numSamples/xwindowwidth
	// which we already have as 1/scaleX
	// so the data to be redrawn starts at n*(1/scaleX) ints
	// and ends at m*(1/scaleX) (n,m being start and end sel)
	// we plot data for pX in this region

  	if ( !XtIsRealized ( _canvas ) )
		return;

	if ( ( _numData == 0 ) || ( startDrawDatumNum == endDrawDatumNum ) )
		return; // no data to plot

	if ( ( startDrawDatumNum > _field2Value ) || ( endDrawDatumNum < _field1Value ) )
		return;
		
  	GC _graphicsContext;
	long pX,pY;
	long startDatumNum;
	long endDatumNum;
	int  centreY, screenX;
	double scaleX,scaleY;
	int oldX, oldY;

	// adjust selection so we draw part or all as dictated by current range
	if ( startDrawDatumNum < _field1Value )
		startDatumNum = _field1Value;
	else
		startDatumNum = startDrawDatumNum;
	
	if ( endDrawDatumNum > _field2Value )
		endDatumNum = _field2Value;
	else
		endDatumNum = endDrawDatumNum;

	getGeometry();
  	_graphicsContext = XCreateGC ( XtDisplay ( _canvas ),
					DefaultRootWindow ( XtDisplay ( _canvas ) ), 0, NULL); 

	scaleX = ( float ) ( _windowWidth - 1 ) / ( float ) ( _field2Value - _field1Value );

	if ( _centreZero == True )
	{
		centreY = _windowHeight / 2;
		scaleY = ( float )  ( _windowHeight ) / ( float ) ( _minY - _maxY );
	}
	else
	{
		centreY = _windowHeight;
		scaleY = - ( float ) _windowHeight / ( float ) ( _maxY - _minY );
	}

	scaleY *= _field7Value;

	// initialise first datum value	
	pY = ( long ) ( ( int ) ( centreY + ( _data[ startDatumNum - _startData ] * scaleY ) ) );
	oldX = ( int ) ( ( startDatumNum - _field1Value ) * scaleX );
	oldY = pY;

	if ( !_selecting )
	{
		if ( _drawing == False )
		{
			// highlight selection background unless drawing
			XSetForeground ( XtDisplay ( _canvas ), _graphicsContext, _pixel[1] );
	
			XFillRectangle ( XtDisplay ( _canvas ), XtWindow ( _canvas ),
								_graphicsContext, oldX, 0, ( ( ( int ) ( ( endDatumNum -
								_field1Value ) * scaleX ) ) - oldX ), _windowHeight );

			XSetForeground ( XtDisplay ( _canvas ), _graphicsContext, _pixel[0] );
		}
		else
		{
			XSetForeground ( XtDisplay ( _canvas ), _graphicsContext, _pixel[0] );
	
			XFillRectangle ( XtDisplay ( _canvas ), XtWindow ( _canvas ),
								_graphicsContext, oldX, 0, ( ( ( int ) ( ( endDatumNum -
								_field1Value ) * scaleX ) ) - oldX ), _windowHeight );

			XSetForeground ( XtDisplay ( _canvas ), _graphicsContext, _pixel[1] );
		}	
	}
	else
		XSetForeground ( XtDisplay ( _canvas ), _graphicsContext, _pixel[1] );

	int increment = 1;	// default not interpolation

	if ( _interpolate == True )
	{
		increment = ( endDatumNum - startDatumNum ) / _windowWidth;
			
		if ( increment == 0 )
			increment = 1;
	}

	for ( pX = startDatumNum; pX <= endDatumNum; pX += increment )
	{
		screenX = ( int ) ( ( pX - _field1Value ) * scaleX );	// x position of screen
		pY = ( long ) ( ( int ) ( centreY + ( _data[ pX - _startData ] * scaleY ) ) );


		if ( _drawPoints == True )
		{
			oldX = screenX;
			oldY = pY;
		}
	
		XDrawLine( XtDisplay ( _canvas ), XtWindow ( _canvas ),
						_graphicsContext, oldX, oldY, screenX, pY );
		
		oldX = screenX;
		oldY = pY;
	}
}


void GraphIntCmd::acquireColour ( int numPixels )
{
	Display *display = XtDisplay ( _canvas );
	Colormap cmap = DefaultColormapOfScreen ( XtScreen ( _canvas ) );

	_colourAcquired = XAllocColorCells ( display, cmap, FALSE, NULL, 0, _pixel, numPixels );
	XtVaGetValues ( XtParent ( _canvas ), XmNbackground, &_pixel[0], NULL );
	XtVaGetValues ( XtParent ( _canvas ), XmNforeground, &_pixel[1], NULL );	// graph highlight colour
	XtVaGetValues ( _canvas, XmNborderColor, &_pixel[2], NULL );	// main drawing colour
}


void GraphIntCmd::setColour ( int r, int g, int b, int pixelNumber )
{
	if ( _colourAcquired )
	{
		Display *display = XtDisplay ( _canvas );
		Colormap cmap = DefaultColormapOfScreen ( XtScreen ( _canvas ) );
		XColor colour;

		_r = (r != -1) ? r : _r;
		_g = (g != -1) ? g : _g;
		_b = (b != -1) ? b : _b;

		colour.red   = _r;
		colour.green = _g;
		colour.blue  = _b;
		colour.flags = DoRed | DoGreen | DoBlue;
		colour.pixel = _pixel[pixelNumber];
		XStoreColor ( display, cmap, &colour );
	}
}


void GraphIntCmd::releaseColour ( int numPixels )
{
	if ( _colourAcquired )
	{
		Display *display = XtDisplay ( _canvas );
		Colormap cmap = DefaultColormapOfScreen ( XtScreen ( _canvas ) );

		XFreeColors ( display, cmap, _pixel, numPixels,
				( long unsigned int ) NULL );

		_colourAcquired = 0;
	}
}


void GraphIntCmd::setZoomLabel ( Boolean zoomStatus )
{
	if ( zoomStatus == False )
	{
		_zoomStatus = True;
		XmString xmstr = XmStringCreateLocalized ( "Zoom out" );
		XtVaSetValues ( _zoom, XmNlabelString, xmstr, XmNrecomputeSize,
		false,  NULL );
		XmStringFree ( xmstr );

		changeField1 ( _field3Value );
		changeField2 ( _field4Value );
	}
	else
	{
		_zoomStatus = False;
		XmString xmstr = XmStringCreateLocalized ( "Zoom in" );
		XtVaSetValues ( _zoom, XmNlabelString, xmstr,
		XmNrecomputeSize, false, NULL );
		XmStringFree ( xmstr );
	}
}


void GraphIntCmd::setSelection ( long startSel, long endSel )
{
	changeField3 ( startSel );
	changeField4 ( endSel );
	
	if ( startSel != endSel )
		XtVaSetValues ( _scrollBar, XmNtroughColor, _pixel[2], NULL );
	else
		XtVaSetValues ( _scrollBar, XmNtroughColor, _pixel[0], NULL );
}


int GraphIntCmd::redisplay ( long start, long end, long selStart, long selEnd,
				long xCurrent, long offset )
{
	// ensure we have the correct data if this is called e.g. from another linked display
	if ( ( start < _field1Value ) || ( end > _field2Value ) || ( xCurrent < _field1Value ) || ( xCurrent > _field2Value ) || ( offset != _startData ) )
		_reload = True;
	else
		_reload = False;

	reload();
	Boolean showArrows;
	
	_startData = offset;
	changeField1 ( start );
	changeField2 ( end );
	setSelection ( selStart, selEnd );

	if ( _numData == _maxData )
		showArrows = False;
	else
		showArrows = True;
		
	if ( ( xCurrent < start ) || ( xCurrent > end ) )
	{
		changeField5 ( start );
		changeField6 ( datumValue ( start - offset ) );
	}
	else
	{
		changeField5 ( xCurrent );
		changeField6 ( datumValue ( xCurrent - offset ) );
	}	

	setScrollBar ( _maxData + _numData - 1,
			_numData * ( ( _maxData + _numData ) / _maxData ),
			_numData - 1, showArrows );

	XtVaSetValues ( _scrollBar,	XmNvalue, _field5Value, NULL );
	refreshView();
	return 0;
}


void GraphIntCmd::setWindowSize ( int width, int height )
{
	_windowWidth = width;
	_windowHeight = height;
}


void GraphIntCmd::setBusyCursor()
{
	_busyCursor = XCreateFontCursor ( XtDisplay ( _popup ), XC_watch );
			
	// install left ptr cursor as normal cursor for window's top-level shell
	XDefineCursor ( XtDisplay ( _popup ), XtWindow ( _popup ), _busyCursor );
}


void GraphIntCmd::setNormalCursor()
{
	_normalCursor = XCreateFontCursor ( XtDisplay ( _popup ), XC_left_ptr );
			
	// install left ptr cursor as normal cursor for window's top-level shell
	XDefineCursor ( XtDisplay ( _popup ), XtWindow ( _popup ), _normalCursor );
}


void GraphIntCmd::setScrollBar ( int maximum, int sliderSize, int pageIncrement, Boolean showArrows )
{
	XtVaSetValues ( _scrollBar,
					XmNmaximum, maximum,
					XmNsliderSize, sliderSize,
					XmNpageIncrement, pageIncrement,
					XmNshowArrows, showArrows,
					NULL );
}


void GraphIntCmd::selRange()
{
	changeField3 ( _field1Value );
	changeField4 ( _field2Value );
	_reload = False;
	reload();
	redrawSelection ( _field3Value, _field4Value );
}
