/*
 FileSelect.cc - originally entitled 'SelectFileCmd'
 modified by Will 18, 2000; will@opaline.demon.co.uk
 Added cancel, help.
  
 Adapted from
 Douglas A. Young, 'Object-oriented Programming with C++ and OSF/Motif'
 2nd ed., Prentice-Hall.
*/


#include "FileSelect.h"
#include "Application.h"
#include <Xm/FileSB.h>
#include <Xm/Protocols.h>


FileSelect::FileSelect ( char *name, int active, FileCallback callback,
			HelpCallback helpCallback, CancelCallback cancelCallback,
			void *clientData ) : NoUndoCmd ( name, active )
{
	_callback = callback;
	_helpCallback = helpCallback;
	_cancelCallback = cancelCallback;
	_clientData = clientData;
}


void FileSelect::doCommand()
{
	Atom  WM_DELETE_WINDOW;
	_fileBrowser = XmCreateFileSelectionDialog ( application->baseWidget(),
									 			 ( char * ) name(), NULL, 0 );
	
	XtAddCallback ( _fileBrowser, XmNokCallback, &FileSelect::selectedFileCallback,
					( XtPointer ) this );
					
	XtAddCallback ( _fileBrowser, XmNhelpCallback, &FileSelect::helpFileCallback,
					( XtPointer ) this );

	XtAddCallback ( _fileBrowser, XmNcancelCallback, &FileSelect::cancelFileCallback,
					( XtPointer ) this );
                
	WM_DELETE_WINDOW = XInternAtom ( XtDisplay ( XtParent (_fileBrowser) ),"WM_DELETE_WINDOW", False );
	XmAddWMProtocols ( XtParent ( _fileBrowser ), &WM_DELETE_WINDOW, 1 );
	XmAddWMProtocolCallback ( XtParent ( _fileBrowser ), WM_DELETE_WINDOW, clientMessageProc, ( XtPointer ) this );

	XtManageChild ( _fileBrowser );
}


void FileSelect::clientMessageProc ( Widget w, XtPointer clientData, XtPointer cbs )
{
	FileSelect *obj = ( FileSelect * ) clientData;

	obj->clientMessage ( ( XmAnyCallbackStruct * ) cbs );
	XtDestroyWidget ( w );
}


void FileSelect::clientMessage ( XmAnyCallbackStruct  *cbs )
{
	if ( cbs->reason != XmCR_CANCEL && cbs->reason != XmCR_OK )
		cancel();
}


void FileSelect::selectedFileCallback ( Widget w, XtPointer clientData, XtPointer callData )
{
	FileSelect *obj = ( FileSelect * ) clientData;
	XmFileSelectionBoxCallbackStruct *cb = ( XmFileSelectionBoxCallbackStruct * ) callData;
						
	char *name = NULL;
	XmString xmstr = cb->value;	// selected file
	int status = 0;
	
	if ( xmstr )	// ensure file was selected
	{
		// extract first char. string matching default character set from
		// compound string
		status = XmStringGetLtoR ( xmstr, XmFONTLIST_DEFAULT_TAG, &name );
		
		if ( status )	// successfully extracted string
			obj->selectedFile ( name );
	}

	XtDestroyWidget ( w );
	// string allocated isn't freed
}


void FileSelect::selectedFile ( char * filename )
{
	if ( _callback )
		_callback ( _clientData, filename );
}


void FileSelect::helpFileCallback ( Widget, XtPointer clientData, XtPointer callData )
{
	FileSelect *obj = ( FileSelect * ) clientData;

	obj->helpFile ( callData );
}


void FileSelect::helpFile ( XtPointer callData )
{
	if ( _helpCallback )
		_helpCallback ( callData );	// disentangle this externally
}


void FileSelect::cancelFileCallback ( Widget w, XtPointer clientData, XtPointer )
{
	FileSelect *obj = ( FileSelect * ) clientData;
	
	obj->cancel();
	XtDestroyWidget ( w );
}


void FileSelect::cancel()
{
	if ( _cancelCallback )
		_cancelCallback ( _clientData );
}
