/*
 FileSelect.h - interactive file selection
 modified by Will 18, 2000; will@opaline.demon.co.uk

 Adapted from
 Douglas A. Young, 'Object-oriented Programming with C++ and OSF/Motif'
 2nd ed., Prentice-Hall
*/


#ifndef FILESELECT_H
#define FILESELECT_H

#include "NoUndoCmd.h"
#include <Xm/Xm.h>


typedef void ( *FileCallback ) ( void * , char * );
typedef void ( *HelpCallback ) ( void * );
typedef void ( *CancelCallback ) ( void * );


class FileSelect : public NoUndoCmd
{
	public:
		FileSelect ( char *, int, FileCallback, HelpCallback, CancelCallback, void * );
		
	protected:
		void doCommand();		// called by base class
		FileCallback _callback;	// function to be called
		HelpCallback _helpCallback;
		CancelCallback _cancelCallback;
		void *_clientData;		// data provided by caller
		Widget _fileBrowser;
		virtual void selectedFile ( char * );
		virtual void helpFile ( XtPointer );
		virtual void cancel ( void );
		
	private:
		// callback for WM 'X'
		static void clientMessageProc ( Widget, XtPointer, XtPointer );

		void clientMessage ( XmAnyCallbackStruct * );
		static void selectedFileCallback ( Widget, XtPointer, XtPointer );
		static void helpFileCallback ( Widget, XtPointer, XtPointer );
		static void cancelFileCallback ( Widget, XtPointer, XtPointer );
};

#endif


/*
  Notes:
  
  The constructor passes its name and active args to its base class, storing the
  callback and client data for later use.
  
  doCommand() creates an XmFileSelectionBox dialog widget as a child of
  LessTifApp's main shell widget, eliminating the need for the class to require
  the widget as a parameter. Theis function registers a callback to be called
  upon fileselection, then displays dialog.

  fileSelectedCallback() is called when the user has chosen a file; it retrieves
  the FileSelect object pointer from the client data and casts the callData arg
  to type XmFileSelectionBoxCallbackStruct, which contains information about
  which file has been chosen. The function XmStringGetLtoR() retrieves the data
  portion of a compound string, necessary since the XmFileSelectionBox widget
  reports the file as a compound string.
  On retrieving the name of the selected file fileSelected() is called and the
  dialog widget destroyed.
*/
